@@ -2,14 +2,17 @@
 #include <unistd.h>
 #include <signal.h>
 #include <time.h>
+#include <corosync/confdb.h>
 #include <netinet/in.h>
+#include <sys/types.h>
+#include <sys/wait.h>
 #include "copyright.cf"
 #include "libcman.h"
 #include "cman_tool.h"
 
 #define DEFAULT_CONFIG_MODULE "xmlconfig"
 
-#define OPTION_STRING		("m:n:v:e:2p:c:r:i:N:t:o:k:F:C:VAPwfqah?Xd::")
+#define OPTION_STRING		("m:n:v:e:2p:c:i:N:t:o:k:F:C:VAPwfqazh?XD::Sd::r::")
 #define OP_JOIN			1
 #define OP_LEAVE		2
 #define OP_EXPECTED		3
@@ -22,7 +25,6 @@
 #define OP_SERVICES		10
 #define OP_DEBUG		11
 
-
 static void print_usage(int subcmd)
 {
 	printf("Usage:\n");
@@ -57,6 +59,10 @@ static void print_usage(int subcmd)
 		printf("  -P               Don't set corosync to realtime priority\n");
 		printf("  -X               Use internal cman defaults for configuration\n");
 		printf("  -A               Don't load openais services\n");
+		printf("  -D<fail|warn|none> What to do about the config. Default (without -D) is to\n");
+		printf("                   validate the config. with -D no validation will be done.\n");
+		printf("                   -Dwarn will print errors but allow the operation to continue.\n");
+		printf("  -z		   Disable stderr debugging output.\n");
 		printf("\n");
 	}
 
@@ -115,7 +121,11 @@ static void print_usage(int subcmd)
 
 	if (!subcmd || subcmd == OP_VERSION) {
 		printf("version\n");
-		printf("  -r <config>      A new config version to set on all members\n");
+		printf("  -r               Reload cluster.conf and update config version.\n");
+		printf("  -D <fail,warn,none> What to do about the config. Default (without -D) is to\n");
+		printf("                   validate the config. with -D no validation will be done. -Dwarn will print errors\n");
+		printf("                   but allow the operation to continue\n");
+		printf("  -S               Don't run ccs_sync to distribute cluster.conf (if appropriate)\n");
 		printf("\n");
 	}
 }
@@ -642,11 +652,107 @@ static void set_votes(commandline_t *comline)
 	cman_finish(h);
 }
 
+static int validate_config(commandline_t *comline, int current_version)
+{
+	struct stat st;
+	char command[PATH_MAX];
+	char validator[PATH_MAX];
+	char ccs_quiet[8];
+	int cmd_res;
+
+	/* Look for ccs_config_validate */
+	snprintf(validator, sizeof(validator), "%s/ccs_config_validate", SBINDIR);
+	if (stat(validator, &st) != 0 || !(st.st_mode & S_IXUSR)) {
+		fprintf(stderr, "Cannot find ccs_config_validate, configuration was not checked but assumed to be OK.\n");
+		return 0;
+	}
+
+	if (comline->verbose > 1) {
+		snprintf(ccs_quiet, sizeof(ccs_quiet), " ");
+	} else {
+		snprintf(ccs_quiet, sizeof(ccs_quiet), "-q");
+	}
+
+	if (current_version) {
+		snprintf(command, sizeof(command), "%s %s -R %d",
+			 validator, ccs_quiet, current_version);
+	} else {
+		snprintf(command, sizeof(command), "%s %s",
+			 validator, ccs_quiet);
+	}
+
+	if (comline->verbose > 1)
+		printf("calling '%s'\n", command);
+
+	cmd_res = system(command);
+
+	return WEXITSTATUS(cmd_res);
+}
+
+/* Here we set the COROSYNC_ variables that might be needed by the corosync
+   configuration modules. We just put them into the environment and leave
+   them for the sub-process to pick up.
+   'config_modules' is returned separately because it's needed internally to
+   and it saves the caller from extracting it all over again.
+   We only return 0 (success) if config_modules is returned as without that
+   the caller can't really do anything at all.
+*/
+static int get_config_variables(commandline_t *comline, char **config_modules)
+{
+	int res;
+	int got_iface = 1;
+	char key_name[1024];
+	size_t key_name_len;
+	char key_value[1024];
+	size_t key_value_len;
+	hdb_handle_t confdb_handle;
+	hdb_handle_t cmanp_handle;
+	confdb_callbacks_t callbacks = {
+		.confdb_key_change_notify_fn = NULL,
+		.confdb_object_create_change_notify_fn = NULL,
+		.confdb_object_delete_change_notify_fn = NULL
+	};
+
+	*config_modules = NULL;
+	res = confdb_initialize (&confdb_handle, &callbacks);
+	if (res != CS_OK)
+		return 0;
+
+	res = confdb_object_find_start(confdb_handle, OBJECT_PARENT_HANDLE);
+	if (res != CS_OK)
+		goto finish;
+
+	res = confdb_object_find(confdb_handle, OBJECT_PARENT_HANDLE, "cman_private", strlen("cman_private"), &cmanp_handle);
+	if (res != CS_OK)
+		goto finish;
+
+	res = confdb_key_iter_start(confdb_handle, cmanp_handle);
+	if (res != CS_OK)
+		goto finish;
+
+	while ( (res = confdb_key_iter(confdb_handle, cmanp_handle, key_name, &key_name_len,
+				       key_value, &key_value_len)) == CS_OK) {
+		key_name[key_name_len] = '\0';
+		key_value[key_value_len] = '\0';
+
+		setenv(key_name, key_value, 1);
+		if (strcmp(key_name, "COROSYNC_DEFAULT_CONFIG_IFACE") == 0) {
+			*config_modules = strdup(key_value);
+			got_iface = 0;
+		}
+	}
+
+finish:
+	confdb_finalize(confdb_handle);
+	return got_iface;
+}
+
 static void version(commandline_t *comline)
 {
 	struct cman_version ver;
 	cman_handle_t h;
 	int result;
+	char *config_modules = NULL;
 
 	h = open_cman_handle(1);
 
@@ -659,10 +765,61 @@ static void version(commandline_t *comline)
 		goto out;
 	}
 
+	if (comline->verbose)
+	        printf("Getting config variables\n");
+	if (get_config_variables(comline, &config_modules))
+	        die("cannot get COROSYNC_DEFAULT_CONFIG_IFACE");
+
+	/* By default we validate the configuration first */
+	if (comline->config_validate_opt != VALIDATE_NONE) {
+
+		if (comline->verbose)
+			printf("Validating configuration\n");
+		result = validate_config(comline, ver.cv_config);
+		if (result == 253)
+			/* Unable to find new config version */
+			die("Unable to retrive the new config version\n");
+		if (result == 254)
+			/* Config regression = fail. */
+			die("Not reloading, config version older or equal the running config");
+		if (result == 255)
+			/* Generic error from ccs_config_validate */
+			die("Not reloading, generic error running ccs_config_validate\n"
+			    "Try re-running with -d options");
+		else if (result && comline->config_validate_opt == VALIDATE_FAIL)
+			die("Not reloading, configuration is not valid");
+	}
+
+	/* We don't bother looking for ccs_sync here - just assume it's in /usr/bin and
+	   that it exists. If this is not true then then user can choose to bypass
+	   the disibution and do it themselves.
+	   Note that ccs_sync might prompt on stderr for passwords the first time it is
+	   run.
+	*/
+	if (strstr(config_modules, "xmlconfig") && !comline->nosync_opt) {
+		if (comline->verbose > 1)
+			printf("calling ccs_sync\n");
+		result = system("/usr/bin/ccs_sync");
+		if (result)
+			die("ccs_sync failed.\nIf you have distributed the config file yourself, try re-running with -S\n");
+	}
+
+	if (comline->verbose)
+		printf("Telling cman the new version number\n");
+
 	ver.cv_config = comline->config_version;
 
-	if ((result = cman_set_version(h, &ver)))
-		die("can't set version: %s", cman_error(errno));
+	result = cman_set_version(h, &ver);
+
+	switch(result) {
+		case 0:
+			if (comline->verbose)
+				printf("Configuration succesfully updated or already running\n");
+		break;
+		default:
+			die("Error loading configuration in corosync/cman");
+		break;
+	}
  out:
 	cman_finish(h);
 }
@@ -752,7 +909,9 @@ static void decode_arguments(int argc, char *argv[], commandline_t *comline)
 {
 	int cont = TRUE;
 	int optchar, i;
+	int suboptchar;
 	int show_help = 0;
+	char buf[16];
 
 	while (cont) {
 		optchar = getopt(argc, argv, OPTION_STRING);
@@ -767,6 +926,34 @@ static void decode_arguments(int argc, char *argv[], commandline_t *comline)
 			comline->addresses_opt = 1;
 			break;
 
+		case 'D':
+			/* Just look at the upper-cased version of the first letter of the argument */
+			if (optarg) {
+				suboptchar = optarg[0] & 0x5F;
+				switch (suboptchar)
+				{
+				case 'F':
+					comline->config_validate_opt = VALIDATE_FAIL;
+					break;
+				case 'W':
+					comline->config_validate_opt = VALIDATE_WARN;
+					break;
+				case 'N':
+					comline->config_validate_opt = VALIDATE_NONE;
+					break;
+				default:
+					die("invalid option to -D, it should be 'force', 'warn' or 'none'\n");
+					break;
+				}
+			}
+			else {
+				comline->config_validate_opt = VALIDATE_NONE;
+			}
+			break;
+		case 'S':
+			comline->nosync_opt = 1;
+			break;
+
 		case 'n':
 		        i = comline->num_nodenames;
 			if (i >= MAX_INTERFACES)
@@ -792,8 +979,13 @@ static void decode_arguments(int argc, char *argv[], commandline_t *comline)
 			break;
 
 		case 'r':
-			comline->config_version = get_int_arg(optchar, optarg);
+			comline->config_version = 0;
 			comline->config_version_opt = TRUE;
+			if (optarg) {
+				fprintf(stderr, "Warning: specifying a "
+					"version for the -r flag is "
+				        "deprecated and no longer used\n");
+			}
 			break;
 
 		case 'v':
@@ -856,6 +1048,10 @@ static void decode_arguments(int argc, char *argv[], commandline_t *comline)
 				comline->verbose = 255;
 			break;
 
+		case 'z':
+			comline->nostderr_debug = 1;
+			break;
+
 		case 'w':
 			comline->wait_opt = TRUE;
 			break;
@@ -941,8 +1137,19 @@ static void decode_arguments(int argc, char *argv[], commandline_t *comline)
 			comline->remove = TRUE;
 		} else if (strcmp(argv[optind], "force") == 0) {
 			comline->force = TRUE;
-		} else
-			die("unknown option %s", argv[optind]);
+		} else {
+			snprintf(buf, sizeof(buf),
+				 "%d", atoi(argv[optind]));
+			if (!strcmp(buf, argv[optind]) &&
+			    (comline->config_version_opt == TRUE) &&
+			     comline->operation == OP_VERSION) {
+				fprintf(stderr, "Warning: specifying a "
+					"version for the -r flag is "
+				        "deprecated and no longer used\n");
+			} else {
+				die("unknown option %s", argv[optind]);
+			}
+		}
 
 		optind++;
 	}
@@ -981,10 +1188,43 @@ static void check_arguments(commandline_t *comline)
 		die("timeout is only appropriate with wait");
 }
 
+
+static void do_join(commandline_t *comline, char *envp[])
+{
+	int ret;
+
+	check_arguments(comline);
+
+	if (comline->timeout) {
+		signal(SIGALRM, sigalarm_handler);
+		alarm(comline->timeout);
+	}
+
+	/* By default we validate the configuration first */
+	if (comline->config_validate_opt != VALIDATE_NONE) {
+
+		if (comline->verbose)
+			printf("Validating configuration\n");
+
+		if (validate_config(comline, 0) &&
+		    comline->config_validate_opt == VALIDATE_FAIL)
+			die("Not joining, configuration is not valid\n");
+	}
+
+	join(comline, envp);
+	if (comline->wait_opt || comline->wait_quorate_opt) {
+		do {
+			ret = cluster_wait(comline);
+			if (ret == ENOTCONN)
+				join(comline, envp);
+
+		} while (ret == ENOTCONN);
+	}
+}
+
 int main(int argc, char *argv[], char *envp[])
 {
 	commandline_t comline;
-	int ret;
 
 	prog_name = argv[0];
 
@@ -994,22 +1234,7 @@ int main(int argc, char *argv[], char *envp[])
 
 	switch (comline.operation) {
 	case OP_JOIN:
-		check_arguments(&comline);
-
-		if (comline.timeout) {
-			signal(SIGALRM, sigalarm_handler);
-			alarm(comline.timeout);
-		}
-
-		join(&comline, envp);
-		if (comline.wait_opt || comline.wait_quorate_opt) {
-			do {
-				ret = cluster_wait(&comline);
-				if (ret == ENOTCONN)
-					join(&comline, envp);
-
-			} while (ret == ENOTCONN);
-		}
+		do_join(&comline, envp);
 		break;
 
 	case OP_LEAVE:
