@@ -19,6 +19,7 @@ static pthread_t query_thread;
 static pthread_mutex_t query_mutex;
 static struct list_head fs_register_list;
 static int kernel_monitor_fd;
+static int daemon_ready;
 
 struct client {
 	int fd;
@@ -587,6 +588,12 @@ static void query_lockspaces(int fd, int max)
 		goto out;
 	}
 
+	/* return error if daemon is not even ready to create lockspaces */
+	if (!daemon_ready) {
+		result = -1;
+		goto out;
+	}
+
 	if (ls_count > max) {
 		result = -E2BIG;
 		ls_count = max;
@@ -995,6 +1002,10 @@ static void loop(void)
 		plock_ci = client_add(rv, process_plocks, NULL);
 	}
 
+	query_lock();
+	daemon_ready = 1;
+	query_unlock();
+
 	for (;;) {
 		rv = poll(pollfd, client_maxi + 1, poll_timeout);
 		if (rv == -1 && errno == EINTR) {
@@ -1274,18 +1285,6 @@ static void read_arguments(int argc, char **argv)
 	}
 }
 
-static void set_oom_adj(int val)
-{
-	FILE *fp;
-
-	fp = fopen("/proc/self/oom_adj", "w");
-	if (!fp)
-		return;
-
-	fprintf(fp, "%i", val);
-	fclose(fp);
-}
-
 static void set_scheduler(void)
 {
 	struct sched_param sched_param;
@@ -1322,7 +1321,6 @@ int main(int argc, char **argv)
 	log_level(LOG_INFO, "dlm_controld %s started", RELEASE_VERSION);
 	signal(SIGTERM, sigterm_handler);
 	set_scheduler();
-	set_oom_adj(-16);
 
 	loop();
 
