@@ -28,12 +28,15 @@
 				  resource class if you delete it from
 				  the configuration */
 #define RF_ENFORCE_TIMEOUTS (1<<9) /** Enforce timeouts for this node */
+#define RF_NON_CRITICAL (1<<10) /** stop this resource if it fails */
+#define RF_QUIESCE	(1<<11) /** don't restart this resource */
 
 
 
 #define RES_STOPPED	(0)
 #define RES_STARTED	(1)
 #define RES_FAILED	(2)
+#define RES_DISABLED	(3)
 
 #ifndef SHAREDIR
 #define SHAREDIR		"/usr/share/rgmanager"
@@ -42,6 +45,7 @@
 #define RESOURCE_ROOTDIR	SHAREDIR
 #define RESOURCE_TREE_ROOT	"/cluster/rm"
 #define RESOURCE_BASE		RESOURCE_TREE_ROOT "/resources"
+#define RESOURCE_DEFAULTS	RESOURCE_TREE_ROOT "/resource-defaults"
 #define RESOURCE_ROOT_FMT 	RESOURCE_TREE_ROOT "/%s[%d]"
 
 #define RESOURCE_MAX_LEVELS	100
@@ -97,7 +101,6 @@ typedef struct _resource {
 	char *	r_name;
 	resource_attr_t *	r_attrs;
 	resource_act_t *	r_actions;
-	time_t	r_started;	/** Time this resource was last started */
 	int	r_flags;
 	int	r_refs;
 	int	r_incarnations;	/** Number of instances running locally */
@@ -111,6 +114,7 @@ typedef struct _rg_node {
 	resource_t	*rn_resource;
 	resource_act_t	*rn_actions;
 	restart_counter_t rn_restart_counter;
+	restart_counter_t rn_failure_counter;
 	int	rn_state; /* State of this instance of rn_resource */
 	int	rn_flags;
 	int	rn_last_status;
@@ -127,6 +131,7 @@ int res_start(resource_node_t **tree, resource_t *res, void *ret);
 int res_stop(resource_node_t **tree, resource_t *res, void *ret);
 int res_status(resource_node_t **tree, resource_t *res, void *ret);
 int res_status_inquiry(resource_node_t **tree, resource_t *res, void *ret);
+int res_convalesce(resource_node_t **tree, resource_t *res, void *ret);
 int res_condstart(resource_node_t **tree, resource_t *res, void *ret);
 int res_condstop(resource_node_t **tree, resource_t *res, void *ret);
 int res_exec(resource_node_t *node, int op, const char *arg, int depth);
@@ -146,14 +151,19 @@ int resource_tree_delta(resource_node_t **, resource_node_t **);
    Load/kill resource rule sets
  */
 int load_resource_rules(const char *rpath, resource_rule_t **rules);
-void print_resource_rule(resource_rule_t *rr);
+int load_resource_defaults(int ccsfd, resource_rule_t **rules);
+void print_resource_rule(FILE *fp, resource_rule_t *rule);
+void print_resource_rules(resource_rule_t **rules);
+void dump_resource_rules(FILE *fp, resource_rule_t **rules);
 void destroy_resource_rules(resource_rule_t **rules);
 
 /*
    Load/kill resource sets
  */
 int load_resources(int ccsfd, resource_t **reslist, resource_rule_t **rulelist);
-void print_resource(resource_t *res);
+void print_resource(FILE *fp, resource_t *res);
+void print_resources(resource_t **reslist);
+void dump_resources(FILE *fp, resource_t **reslist);
 void destroy_resources(resource_t **list);
 
 /*
@@ -162,9 +172,12 @@ void destroy_resources(resource_t **list);
 int build_resource_tree(int ccsfd, resource_node_t **tree,
 			resource_rule_t **rulelist, resource_t **reslist);
 void print_resource_tree(resource_node_t **tree);
+void dump_resource_tree(FILE *fp, resource_node_t **tree);
 void destroy_resource_tree(resource_node_t **tree);
 
 void *act_dup(resource_act_t *acts);
+void dump_resource_info(FILE *fp);
+
 
 /*
    Handy functions
